#include <vector>
#include <iostream>
#include <iomanip>

struct FiniteElement
{
    unsigned int elementId;
    unsigned int entityId;
    std::vector<unsigned int> nodesId;

    bool operator<(const FiniteElement& p_el) const {
        return elementId < p_el.elementId;
    }
};

struct Node
{
    unsigned int nodeId;
    bool isEdge;
    std::vector<double> cords;
    Node() = default;
    Node(int id, double x, double y, double z, bool p_isEdge) :
        nodeId(id), cords{x,y,z}, isEdge(p_isEdge) {}

    bool operator<(const Node& p_node) const {
        return nodeId < p_node.nodeId;
    }
};

struct BoundaryFiniteElement : public FiniteElement
{
    unsigned int boundaryID;
};

struct Edge {
    int firstNodeId;
    int lastNodeId;
    int centerNodeId;

    Edge() = default;
    Edge(int id1, int id2, int id3) :
        firstNodeId(id1), lastNodeId(id2), centerNodeId(id3) {}


    bool operator==(const Edge& rightEdge) const {
        if (rightEdge.firstNodeId == firstNodeId && rightEdge.lastNodeId == lastNodeId)
            return true;
        else if (rightEdge.firstNodeId == lastNodeId && rightEdge.lastNodeId == firstNodeId)
            return true;
        else return false;
    }

    void updateCenter(int id) {
        centerNodeId = id;
    }

};


inline std::ostream& operator<<(std::ostream& stream, const Node& node)
{
    stream << "(" << node.nodeId << ") ";
    for (auto& i : node.cords)
    {
        stream << i << " ";
    }
    if (node.isEdge)
    {
        stream << "Is edge";
    }
    else
        stream << "Isnt edge";
    stream << std::endl;
    return stream;
}

inline std::ostream& operator<<(std::ostream& stream, const FiniteElement& finiteElement)
{
    stream << "(" << finiteElement.elementId << ") ";
    stream << "<" << finiteElement.entityId << "> ";
    for (auto& j : finiteElement.nodesId)
    {
        stream << j << " ";
    }
    stream << std::endl;
    return stream;
}

inline std::ostream& operator<<(std::ostream& stream, const BoundaryFiniteElement& boundaryFiniteElement)
{
    stream << "(" << boundaryFiniteElement.elementId << ") ";
    stream << "<" << boundaryFiniteElement.boundaryID << "> ";
    for (auto& j : boundaryFiniteElement.nodesId)
    {
        stream << j << " ";
    }
    stream << std::endl;
    return stream;
}


template < typename _T >
inline void hashCombine(std::size_t & o_seed, const _T & p_val)
{
    o_seed ^= std::hash <_T >() (p_val) + 0x9e3779b9 + (o_seed << 6) + (o_seed >> 2);
    }
template < typename _T >
inline void hashVal(std::size_t & o_seed, const _T & p_val)
{
    hashCombine(o_seed, p_val);
    }

namespace std {
    template<>
    class hash<Edge> {
    public:
        size_t operator()(const Edge& s) const
        {

            int left(s.firstNodeId);
            int right(s.lastNodeId);
            size_t seed = 0;

            if (left > right) {
                swap(left, right);
            }

            hashCombine(seed, left);
            hashCombine(seed, right);

            return seed;
        }
    };
}